from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt

def generate_synthetic_data(n_dim: int, n_clusters: int, n_samples: int) -> tuple:
    """
    Function to generate synthetic data to test the algorithms.

    Return:
        tuple: containing the set of training and test values, with their associated labels
    """

    X, y = make_blobs(n_samples=n_samples, 
                    n_features=n_dim, 
                    centers=n_clusters)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.2)

    return train_X, train_y, test_X, test_y


import numpy as np
from qiskit import QuantumRegister, QuantumCircuit

from .encoding import Encoding
from ..gates import RBS


class UnaryLoader(Encoding):
    """
    Encoding of classical data into quantum info based on a unary basis, i.e., |10..0>, 
    |01..0>, ..., |00..1>.

    With this encoding we obtain d qubits, d-1 parametrized two qubits gates (plus a cnot
    in the first qubit) and a log d depth of the circuit.

    With this codification for two data points, if we invert one of them, connect it to the
    other and measure the probability of reading |1> in the first qubit, we obtain the inner 
    product of the two data points normalized.
    """

    def __init__(self, input_vector: np.ndarray, inverse: bool = False) -> None:
        """

        Args:
            input_vector: Vector of k dimensions
            inverse: Boolean that determines whether we should invert the circuit or not
                    (in order to obtain the inner product)
        """
        self.num_qubits = int(len(input_vector))
        self.quantum_data = QuantumRegister(self.num_qubits)
        self.circuit = QuantumCircuit(self.quantum_data)
        newx = np.copy(input_vector)

        betas = []
        self._theta_calc(newx, betas)
        self._generate_circuit(betas)

        super().__init__("Unary encoding")

    def _generate_circuit(self, betas: np.ndarray) -> None:
        """
        Function to generate the circuit with the info of the real data.

        Args:
            betas: Vector of k-1 dimensions with the angles for the gates 
        """
        logarithm = int(np.log2(self.num_qubits))

        # Generation a cicuit with a depth of log d, being d the number of qubits, using the RBS gate 
        for i in range(logarithm):
            for k in range(2**i):
                w = self.num_qubits // 2**i
                self.circuit.unitary(RBS(betas[2**i + k - 1]).rbs, [self.quantum_data[k*w],
                                                                    self.quantum_data[k*w + w // 2]], label="RBS({})".format(np.around(betas[2**i + k - 1], 3)))

    def _theta_calc(self, input_vector: np.ndarray, betas: np.ndarray) -> None:
        """
        Function to calculate the angles for the gates of the circuit. We follow the procedure of the paper [1] in
        the section Methods.


        ** References: **
        [1] Sonika Johri, Shantanu Debnath, Avinash Mocherla, Alexandros SINGK, Anupam Prakash, Jungsang Kim 
        and Iordanis Kerenidis et al.,
        `Nearest centroid classification on a trapped ion quantum computer 
        <https://www.nature.com/articles/s41534-021-00456-5>`
        Args:
            input_vector: Vector of k dimensions to be encoded.
        """
        d_half = len(input_vector)//2
        rigth_half = input_vector[d_half:]

        # r = (r_1, ..., r_(d-1))
        r = []

        # Calculations for r_(d/2), ..., r_(d-1)
        for i in range(len(rigth_half)):
            r_elem = np.sqrt(input_vector[2*i]**2 + input_vector[2*i+1]**2)
            r.append(r_elem)

        # Calculations for r_1, ..., r_(d/2)
        for i in range(d_half-2, -1, -1):
            r_elem = np.sqrt(r[i+1]**2 + r[i]**2)
            r.insert(0, r_elem)

        # Calculation of the θ's
        for i in range(len(input_vector)-1):

            # Calculations for θ_1, ..., θ_(d/2)
            if i < d_half - 1:
                if r[i] != 0:
                    betas.append(np.arccos(r[2*i+1] / r[i]))
                else:
                    betas.append(1)

            # Claculations for θ_(d/2), ..., θ_d
            else:
                if input_vector[(i - d_half + 1)*2 + 1] >= 0:
                    if r[i] != 0:
                        betas.append(
                            np.arccos(input_vector[(i - d_half + 1)*2] / r[i]))
                    else:
                        betas.append(1)
                else:
                    if r[i] != 0:
                        betas.append(
                            2*np.pi - np.arccos(input_vector[(i - d_half + 1)*2] / r[i]))
                    else:
                        betas.append(1)



import numpy as np

from qiskit import QuantumCircuit

class Encoding:
    """
    Base class for any encoding of classical data into quantum data.
    """
    def __init__(self, name: str) -> None:
        """
        Args:
            name: Name of the encoding.
        """
        self.name = name

    @property
    def circuit(self) -> QuantumCircuit:
        return self._circuit

    @circuit.setter
    def circuit(self, circuit: QuantumCircuit) -> None:
        self._circuit = circuit


    def _generate_circuit(self, input_vector: np.ndarray):
        """
        Function to generate the circuit with the info of the real data.

        Args:
            input_vector: Real data
        """
        raise NotImplementedError()



import numpy as np

from qiskit import QuantumRegister, QuantumCircuit
from qiskit.circuit.gate import Gate
from qiskit.circuit.library import CRYGate, CRZGate, CXGate
from qiskit.circuit.parameterexpression import ParameterValueType
from typing import Optional


class iRBS(Gate):
    r"""The SWAP gate.
    This is a symmetric and Clifford gate.
    **Circuit symbol:**
    .. parsed-literal::
        q_0: ─X─
              │
        q_1: ─X─
    **Matrix Representation:**
    .. math::
        SWAP =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1
            \end{pmatrix}
    The gate is equivalent to a state swap and is a classical logic gate.
    .. math::
        |a, b\rangle \rightarrow |b, a\rangle
    """

    def __init__(self, theta: ParameterValueType, label: Optional[str] = None):
        """Create new iRBS gate."""
        super().__init__("iRBS", 2, [theta])

    def _define(self):
        # pylint: disable=cyclic-import

        #                 ┌───┐┌──────────────┐                     ┌───┐┌──────────┐
        # q_0: ───────────┤ X ├┤   Rz(-π/2)   ├──■──────────────────┤ X ├┤ Rz(-π/2) ├
        #      ┌─────────┐└─┬─┘├──────────────┤┌─┴─┐┌──────────────┐└─┬─┘└──────────┘
        # q_1: ┤ Rz(π/2) ├──■──┤ Ry(π/2 - 2ϴ) ├┤ X ├┤ Ry(2ϴ - π/2) ├──■──────────────
        #      └─────────┘     └──────────────┘└───┘└──────────────┘

        q = QuantumRegister(2, "q")
        theta = self.params[0]
        qc = QuantumCircuit(q, name=self.name)
        rules = [
            (CRZGate(np.pi / 2), [q[1]], []),
            (CXGate(), [q[1], q[0]], []),
            (CRZGate(-np.pi / 2), [q[0]], []),
            (CRYGate(np.pi / 2 - 2*theta), [q[1]], []),
            (CXGate(), [q[0], q[1]], []),
            (CRYGate(2*theta - np.pi / 2 ), [q[1]], []),
            (CXGate(), [q[1], q[0]], []),
            (CRZGate(-np.pi / 2), [q[0]], []),
        ]
        for instr, qargs, cargs in rules:
            qc._append(instr, qargs, cargs)

        self.definition = qc

    def get_iRBS(self):
        return self.definition


import numpy as np
from qiskit import Aer
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister

from ..encoding import UnaryLoader


class NearestCentroid:
    """Nearest Centroid classifier.

    The Nearest Centroid is a classification algorithm that assigns each observation a label
    based on which centroid of the different clusters available in the data is the closest.

    For doing this, we load the classical data to quantum info using a Unary Loader [1] and calculate
    an inner product of the vectors. We use this inner product to estimate the distance between
    the observation with each centroid to, only then, select the closest to assign a label to it.

    ** References: **
    [1] Sonika Johri, Shantanu Debnath, Avinash Mocherla, Alexandros SINGK, Anupam Prakash, Jungsang Kim 
    and Iordanis Kerenidis et al.,
    `Nearest centroid classification on a trapped ion quantum computer 
    <https://www.nature.com/articles/s41534-021-00456-5>`

    """

    def __init__(self, X: np.ndarray = None, y: np.ndarray = None, n_dim: int = None) -> None:
        """

        Args:
            X: Training data values of dimension k or None
            y: Training data labels of dimension k or None
            n_dim: Dimension of the training dataset, i.e., k or None
        Raise: 
            AttributeError: In case some attributes have values and others are empty
        """
        if (X is not None) and (y is not None) and (n_dim is not None):
            self.train_X = X
            self.train_y = y
            self.n_dim = n_dim
        elif (X is None) and (y is None) and (n_dim is None):
            pass
        else:
            raise AttributeError("Either all attributes are empty or have the accurate values.")
    
    def _calc_circ_centroids(self) -> None:
        """ Calculate centroids' circuits """
        self.centroids = self._calc_centroids()

        self.circ_centroids = {}
        for label, centroid in self.centroids.items():
            
            # After we calculate each centroid, we create its respective circuit
            qregisters = QuantumRegister(self.n_dim, "q")
            cregisters = ClassicalRegister(1, "c")
            circ_centroid = QuantumCircuit(qregisters, cregisters)
            circ_centroid.x(qregisters[0])
            circ_centroid.compose(UnaryLoader(centroid).circuit, inplace=True)
            self.circ_centroids[label] = circ_centroid
    
    def _calc_centroids(self) -> dict:
        """ 
        Calculate the centroids of the clusters classically

        Returns:
            dict: with the centroid point associated to each label
        """
        centroids = dict.fromkeys(self.train_y)
        for label in centroids:
            centroids[label] = np.repeat(float(0), self.n_dim)
        for x, label in zip(self.train_X, self.train_y):
            centroids[label] += x / len(self.train_X)
        return centroids

    def _quantum_distance(self, circ_centroid: QuantumCircuit, centroid: np.ndarray, y: np.ndarray, repetitions: int = 1000) -> float:
        """
        Calculates the distance between x (centroid) and y using circuit to calculate the inner product. The distance is the
        euclidean distance, represented by the formula:

        dist = sqrt(||x||**2 + ||y||**2 - 2*||x||*||y||*inner)

        where inner = <x|y>, i.e., inner product between x and y.

        Args:
            circ_centroid: Unary loader of the centroid
            centroid: First vector to perform the distance
            y: First vector to perform the distance
        Returns:
            float: distance between x and y
        """
        norm_x = np.linalg.norm(centroid)
        norm_y = np.linalg.norm(y)
        inner = self._quantum_inner(circ_centroid, y, repetitions)
        dist = np.sqrt(norm_x**2 + norm_y**2 - 2 * norm_x * norm_y * inner)
        return dist

    def _quantum_inner(self, circ_centroid: QuantumCircuit, y: np.ndarray, repetitions: int = 1000) -> float:
        """
        Inner product calculation of x (centroid) and y.

        Args:
            circ_centroid: Unary loader of the centroid
            y: First vector to perform the distance
        Returns:
            float: inner product between a centroid and y
        """ 
        # We execute the circuit a defined times of repetitions
        circ_y = UnaryLoader(y, inverse=True).circuit.inverse()
        circuit = circ_centroid.compose(circ_y)
        circuit.measure([0], [0])
        simulator = Aer.get_backend('aer_simulator')
        result = simulator.run(circuit, shots=repetitions).result()
        counts = result.get_counts(circuit)

        if '1' in counts.keys():
            z = counts['1'] / repetitions
        else:
            z = 0
        return np.sqrt(z)

    def fit(self, X: np.ndarray = None, y: np.ndarray = None, n_dim: int = None) -> None:
        if (X is not None) and (y is not None):
            self.train_X = X
            self.train_y = y
            self.n_dim = n_dim
            self._calc_circ_centroids()
        elif (X is None) and (y is None):
            if (self.train_X is not None) and (self.train_y is not None):
                self._calc_circ_centroids()
            else:
                raise AttributeError("You need to initialize the dataset.")

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Predicts the label of a batch of observations

        Args:
            X: Test data values of dimension l
            y: Test data labels of dimension l
        Returns:
            nd.ndarray: with the label of each observation in the batch
        """
        if self.circ_centroids is not None:
            labels = []
            for x in X:
                dist = {}
                # For every x in the test set X we calculate the distance to each centroid and select the minimum 
                for label, centroid in self.centroids.items():
                    dist[label] = self._quantum_distance(
                        self.circ_centroids[label], centroid, x)
                labels.append(min(dist, key=dist.get))
            return np.array(labels)
        else:
            raise AttributeError("Training set cannot be empty to predict.")

    


import numpy as np
import matplotlib.pyplot as plt

def error_graph(y_test: np.ndarray, img_name: str, title:str, *args):
    names = []
    accuracy = []
    for prediction in args:
        names.append(prediction[0])
        counts = 0
        for i, y in enumerate(prediction[1]):
            if y == y_test[i]:
                counts += 1
        accuracy.append(counts / len(y_test))

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_ylabel('Accuracy')
    ax.set_xlabel('Method used')
    ax.set_title(title)
    ax.bar(names,accuracy)
    plt.savefig(img_name)

def variance_error_graph(y_test: np.ndarray, img_name: str, *args):
    names = []
    accuracy = []
    CTEs = []
    error = []
    for prediction in args:
        names.append(prediction[0])
        for iteration in prediction[1]:
            counts = 0
            for i, y in enumerate(iteration):
                if y == y_test[i]:
                    counts += 1
            accuracy.append(counts / len(y_test))
        CTEs.append(np.mean(accuracy))
        error.append(np.std(accuracy))

    x_pos = np.arange(len(args))

    # Build the plot
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.bar(x_pos, CTEs, yerr=error, align='center', alpha=0.5, ecolor='black', capsize=10)
    for i,j in zip(x_pos, CTEs):
        ax.annotate(str(j),xy=(i,j))
    ax.set_ylabel('Accuracy')
    ax.set_xticks(x_pos)
    ax.set_xticklabels(names)
    ax.yaxis.grid(True)

    # Save the figure and show
    plt.tight_layout()
    plt.savefig(img_name)
    plt.show()
    
